@@ -51,6 +51,7 @@ extern "C" {
 #include <kiwix/manager.h>
 #include <kiwix/reader.h>
 #include <kiwix/searcher.h>
+#include <kiwix/opds_dumper.h>
 #include <pthread.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -81,6 +82,8 @@ extern "C" {
 #include <unistd.h>
 #endif
 
+#include "request_context.h"
+
 #ifdef interface
 #undef interface
 #endif
@@ -90,12 +93,14 @@ using namespace std;
 static bool noLibraryButtonFlag = false;
 static bool noSearchBarFlag = false;
 static string welcomeHTML;
+static string catalogOpenSearchDescription;
 static std::atomic_bool isVerbose(false);
 static std::string rootLocation = "";
 static std::map<std::string, std::string> extMimeTypes;
 static std::map<std::string, kiwix::Reader*> readers;
 static std::map<std::string, kiwix::Searcher*> searchers;
 static kiwix::Searcher* globalSearcher = nullptr;
+static kiwix::Manager libraryManager;
 static pthread_mutex_t searchLock = PTHREAD_MUTEX_INITIALIZER;
 static pthread_mutex_t compressorLock = PTHREAD_MUTEX_INITIALIZER;
 static pthread_mutex_t regexLock = PTHREAD_MUTEX_INITIALIZER;
@@ -124,35 +129,11 @@ static std::string getMimeTypeForFile(const std::string& filename)
 }
 
 
-struct RequestContext {
-    struct MHD_Connection* connection;
-    int httpResponseCode;
-    kiwix::Reader* reader;
-    kiwix::Searcher* searcher;
-    const std::string urlStr;
-    const std::string humanReadableBookId;
-    bool acceptEncodingDeflate;
-    bool acceptRange;
-    int range_start;
-    int range_end;
-
-    RequestContext(struct MHD_Connection* connection, int httpResponseCode,
-                   kiwix::Reader* reader, kiwix::Searcher* searcher,
-                   const std::string& urlStr, const std::string& humanReadableBookId,
-                   bool acceptEncodingDeflate,
-                   bool acceptRange, int range_start, int range_end) :
-        connection(connection),
-        httpResponseCode(httpResponseCode),
-        reader(reader),
-        searcher(searcher),
-        urlStr(urlStr),
-        humanReadableBookId(humanReadableBookId),
-        acceptEncodingDeflate(acceptEncodingDeflate),
-        acceptRange(acceptRange),
-        range_start(range_start),
-        range_end(range_end)
-    {}
-};
+static bool startswith(const std::string& base, const std::string& start)
+{
+   return start.length() <= base.length()
+        && std::equal(start.begin(), start.end(), base.begin());
+}
 
 
 void introduceTaskbar(string& content, const string& humanReadableBookId)
@@ -162,9 +143,7 @@ void introduceTaskbar(string& content, const string& humanReadableBookId)
     content = appendToFirstOccurence(
         content,
         "<head>",
-        replaceRegex(
-            RESOURCE::include_html_part, humanReadableBookId, "__CONTENT__")
-        + (noLibraryButtonFlag
+            RESOURCE::include_html_part + (noLibraryButtonFlag
             ? "<style>#kiwix_serve_taskbar_library_button { display: none }</style>"
             : "")
     );
@@ -183,6 +162,7 @@ void introduceTaskbar(string& content, const string& humanReadableBookId)
              "__CONTENT__"));
     }
     content = replaceRegex(content, rootLocation, "__ROOT_LOCATION__");
+    content = replaceRegex(content, replaceRegex(humanReadableBookId, "%26", "&"), "__CONTENT_ESCAPED__");
   }
   pthread_mutex_unlock(&regexLock);
 }
@@ -241,8 +221,8 @@ static struct MHD_Response* build_response(const void* data,
                                            bool cacheEnabled)
 {
   /* Create the response */
-  struct MHD_Response* response = MHD_create_response_from_data(
-      length, const_cast<void*>(data), MHD_NO, MHD_YES);
+  struct MHD_Response* response = MHD_create_response_from_buffer(
+      length, const_cast<void*>(data), MHD_RESPMEM_MUST_COPY);
 
   /* Make a redirection if necessary otherwise send the content */
   if (!httpRedirection.empty()) {
@@ -286,39 +266,47 @@ static struct MHD_Response* build_response(const void* data,
 }
 
 
-static struct MHD_Response* build_404(RequestContext* request_context) {
+static struct MHD_Response* build_404(RequestContext* request,
+                                      const std::string& humanReadableBookId) {
     std::string content
         = "<!DOCTYPE html>\n<html><head><meta "
           "content=\"text/html;charset=UTF-8\" http-equiv=\"content-type\" "
           "/><title>Content not found</title></head><body><h1>Not "
           "Found</h1><p>The requested URL \""
-          + request_context->urlStr + "\" was not found on this server.</p></body></html>";
+          + request->get_full_url() + "\" was not found on this server.</p></body></html>";
     auto mimeType = "text/html";
-    request_context->httpResponseCode = MHD_HTTP_NOT_FOUND;
-    introduceTaskbar(content, request_context->humanReadableBookId);
+    request->httpResponseCode = MHD_HTTP_NOT_FOUND;
+    introduceTaskbar(content, humanReadableBookId);
     bool deflated
-        = request_context->acceptEncodingDeflate && compress_content(content, mimeType);
+        = request->can_compress() && compress_content(content, mimeType);
     return build_response(
         content.data(), content.size(), "", mimeType, deflated, false);
 }
 
+static struct MHD_Response* build_homepage(RequestContext* request)
+{
+  std::string content = welcomeHTML;
+
+  std::string mimeType = "text/html; charset=utf-8";
+
+  bool deflated = request->can_compress() && compress_content(content, mimeType);
+  return build_response(
+      content.data(), content.size(), "", mimeType, deflated, false);
+}
+
 struct RunningResponse {
-   zim::Article* article;
+   kiwix::Entry entry;
    int range_start;
 
-   RunningResponse(zim::Article* article,
+   RunningResponse(kiwix::Entry entry,
                    int range_start) :
-     article(article),
+     entry(entry),
      range_start(range_start)
    {}
-
-   ~RunningResponse() {
-     delete article;
-   }
 };
 
 
-ssize_t callback_reader_from_article(void* cls,
+ssize_t callback_reader_from_entry(void* cls,
                                   uint64_t pos,
                                   char* buf,
                                   size_t max)
@@ -327,39 +315,39 @@ ssize_t callback_reader_from_article(void* cls,
 
   size_t max_size_to_set = min<size_t>(
     max,
-    response->article->getArticleSize() - pos - response->range_start);
+    response->entry.getSize() - pos - response->range_start);
 
   if (max_size_to_set <= 0) {
     return MHD_CONTENT_READER_END_WITH_ERROR;
   }
 
-  zim::Blob blob = response->article->getData(response->range_start+pos, max_size_to_set);
+  zim::Blob blob = response->entry.getBlob(response->range_start+pos, max_size_to_set);
   memcpy(buf, blob.data(), max_size_to_set);
   return max_size_to_set;
 }
 
-void callback_free_article(void* cls)
+void callback_free_response(void* cls)
 {
   RunningResponse* response = static_cast<RunningResponse*>(cls);
   delete response;
 }
 
-static struct MHD_Response* build_callback_response_from_article(
-    zim::Article& article, int range_start, int range_len, const std::string& mimeType)
+static struct MHD_Response* build_callback_response_from_entry(
+   kiwix::Entry entry, int range_start, int range_len, const std::string& mimeType)
 {
   RunningResponse* run_response =
-      new RunningResponse(new zim::Article(article), range_start);
+      new RunningResponse(entry, range_start);
 
   struct MHD_Response* response
-      = MHD_create_response_from_callback(article.getArticleSize(),
+      = MHD_create_response_from_callback(entry.getSize(),
                                           16384,
-                                          callback_reader_from_article,
+                                          callback_reader_from_entry,
                                           run_response,
-                                          callback_free_article);
+                                          callback_free_response);
   /* Tell the client that byte ranges are accepted */
   MHD_add_response_header(response, MHD_HTTP_HEADER_ACCEPT_RANGES, "bytes");
   std::ostringstream oss;
-  oss << "bytes " << range_start << "-" << range_start + range_len - 1 << "/" << article.getArticleSize();
+  oss << "bytes " << range_start << "-" << range_start + range_len - 1 << "/" << entry.getSize();
 
   MHD_add_response_header(response, MHD_HTTP_HEADER_CONTENT_RANGE, oss.str().c_str());
 
@@ -381,29 +369,98 @@ static struct MHD_Response* build_callback_response_from_article(
   return response;
 }
 
-static struct MHD_Response* handle_suggest(RequestContext* request_context)
+std::pair<kiwix::Reader*, kiwix::Searcher*>
+get_from_humanReadableBookId(const std::string& humanReadableBookId) {
+  kiwix::Searcher* searcher
+        = searchers.find(humanReadableBookId) != searchers.end()
+              ? searchers.find(humanReadableBookId)->second
+              : globalSearcher;
+  kiwix::Reader* reader = readers.find(humanReadableBookId) != readers.end()
+                                ? readers.find(humanReadableBookId)->second
+                                : NULL;
+  return std::pair<kiwix::Reader*, kiwix::Searcher*>(reader, searcher);
+}
+
+static struct MHD_Response* handle_meta(RequestContext* request)
 {
+  std::string humanReadableBookId;
+  std::string meta_name;
+  try {
+    humanReadableBookId = request->get_argument("content");
+    meta_name = request->get_argument("name");
+  } catch (const std::out_of_range& e) {
+    return build_404(request, humanReadableBookId);
+  }
+
+  auto reader = get_from_humanReadableBookId(humanReadableBookId).first;
+  if (reader == nullptr) {
+    return build_404(request, humanReadableBookId);
+  }
+
+  std::string content;
+  std::string mimeType = "text";
+
+  if (meta_name == "title") {
+    content = reader->getTitle();
+  } else if (meta_name == "description") {
+    content = reader->getDescription();
+  } else if (meta_name == "language") {
+    content = reader->getLanguage();
+  } else if (meta_name == "name") {
+    content = reader->getName();
+  } else if (meta_name == "tags") {
+    content = reader->getTags();
+  } else if (meta_name == "date") {
+    content = reader->getDate();
+  } else if (meta_name == "creator") {
+    content = reader->getCreator();
+  } else if (meta_name == "publisher") {
+    content = reader->getPublisher();
+  } else if (meta_name == "favicon") {
+    reader->getFavicon(content, mimeType);
+  } else {
+    return build_404(request, humanReadableBookId);
+  }
+
+  return build_response(content.data(), content.size(), "", mimeType, false, true);
+}
+
+static struct MHD_Response* handle_suggest(RequestContext* request)
+{
+  if (isVerbose.load()) {
+    printf("** running handle_suggest\n");
+  }
+
   std::string content;
   std::string mimeType;
   unsigned int maxSuggestionCount = 10;
   unsigned int suggestionCount = 0;
   std::string suggestion;
 
-  /* Get the suggestion pattern from the HTTP request */
-  const char* cTerm = MHD_lookup_connection_value(
-    request_context->connection, MHD_GET_ARGUMENT_KIND, "term");
-  std::string term = cTerm == NULL ? "" : cTerm;
+  std::string humanReadableBookId;
+  std::string term;
+  try {
+    humanReadableBookId = request->get_argument("content");
+    term = request->get_argument("term");
+  } catch (const std::out_of_range&) {
+    return build_homepage(request);
+  }
+
   if (isVerbose.load()) {
     printf("Searching suggestions for: \"%s\"\n", term.c_str());
   }
 
+  auto reader_searcher = get_from_humanReadableBookId(humanReadableBookId);
+  auto reader = reader_searcher.first;
+  auto searcher = reader_searcher.second;
+
   pthread_mutex_lock(&searchLock);
   /* Get the suggestions */
   content = "[";
-  if (request_context->reader != NULL) {
+  if (reader != nullptr) {
     /* Get the suggestions */
-    request_context->reader->searchSuggestionsSmart(term, maxSuggestionCount);
-    while (request_context->reader->getNextSuggestion(suggestion)) {
+    reader->searchSuggestionsSmart(term, maxSuggestionCount);
+    while (reader->getNextSuggestion(suggestion)) {
       kiwix::stringReplacement(suggestion, "\"", "\\\"");
       content += (content == "[" ? "" : ",");
       content += "{\"value\":\"" + suggestion + "\",\"label\":\"" + suggestion
@@ -414,225 +471,343 @@ static struct MHD_Response* handle_suggest(RequestContext* request_context)
   pthread_mutex_unlock(&searchLock);
 
   /* Propose the fulltext search if possible */
-  if (request_context->searcher != NULL) {
+  if (searcher != NULL) {
     content += (suggestionCount == 0 ? "" : ",");
-    content += "{\"value\":\"" + std::string(term)
-               + " \", \"label\":\"containing '" + std::string(term)
+    content += "{\"value\":\"" + term
+               + " \", \"label\":\"containing '" + term
                + "'...\"}";
   }
 
   content += "]";
   mimeType = "application/json; charset=utf-8";
-  bool deflated = request_context->acceptEncodingDeflate && compress_content(content, mimeType);
+  bool deflated = request->can_compress() && compress_content(content, mimeType);
   return build_response(
       content.data(), content.size(), "", mimeType, deflated, true);
 }
 
-static struct MHD_Response* handle_skin(RequestContext* request_context)
+static struct MHD_Response* handle_skin(RequestContext* request)
 {
-  std::string content;
-  try {
-    content = getResource(request_context->urlStr.substr(rootLocation.size() + 6));
-  } catch (const ResourceNotFound& e) {
-    return build_404(request_context);
+  if (isVerbose.load()) {
+    printf("** running handle_skin\n");
   }
-  std::string mimeType = getMimeTypeForFile(request_context->urlStr);
-  bool deflated = request_context->acceptEncodingDeflate && compress_content(content, mimeType);
+
+  std::string content;
+  auto resourceName = request->get_url().substr(6);
+  try {
+    content = getResource(resourceName);
+  } catch (const ResourceNotFound& e) {
+    return build_404(request, "");
+  }
+  std::string mimeType = getMimeTypeForFile(resourceName);
+  bool deflated = request->can_compress() && compress_content(content, mimeType);
   return build_response(
       content.data(), content.size(), "", mimeType, deflated, true);
 }
 
-static struct MHD_Response* handle_search(RequestContext* request_context)
+static struct MHD_Response* handle_search(RequestContext* request)
 {
+  if (isVerbose.load()) {
+    printf("** running handle_search\n");
+  }
+
   std::string content;
   std::string mimeType;
   std::string httpRedirection;
 
-  /* Retrieve the pattern to search */
-  const char* pattern = MHD_lookup_connection_value(
-      request_context->connection, MHD_GET_ARGUMENT_KIND, "pattern");
-  std::string patternString
-      = kiwix::urlDecode(pattern == NULL ? "" : string(pattern));
-  std::string patternCorrespondingUrl;
+  std::string humanReadableBookId;
+  std::string patternString;
+  try {
+    humanReadableBookId = request->get_argument("content");
+  } catch (const std::out_of_range&) {}
+
+  try {
+    patternString = request->get_argument("pattern");
+  } catch (const std::out_of_range&) {}
+
+  /* Retrive geo search */
+  bool has_geo_query = false;
+  float latitude;
+  float longitude;
+  float distance;
+  try {
+    latitude = request->get_argument<float>("latitude");
+    longitude = request->get_argument<float>("longitude");
+    distance = request->get_argument<float>("distance");
+    has_geo_query = true;
+  } catch(const std::out_of_range&) {}
+    catch(const std::invalid_argument&) {}
+
+  /* Search results for searches from the welcome page should not
+     be cached
+  */
+  auto reader_searcher = get_from_humanReadableBookId(humanReadableBookId);
+  auto reader = reader_searcher.first;
+  auto searcher = reader_searcher.second;
+  bool cacheEnabled = !(searcher == globalSearcher);
 
   /* Try first to load directly the article */
-  if (request_context->reader != NULL) {
-    std::vector<std::string> variants = request_context->reader->getTitleVariants(patternString);
-    std::vector<std::string>::iterator variantsItr = variants.begin();
+  if (reader != nullptr && !patternString.empty()) {
+    std::string patternCorrespondingUrl;
+    auto variants = reader->getTitleVariants(patternString);
+    auto variantsItr = variants.begin();
 
     while (patternCorrespondingUrl.empty() && variantsItr != variants.end()) {
-      request_context->reader->getPageUrlFromTitle(*variantsItr, patternCorrespondingUrl);
-      variantsItr++;
+      try {
+        auto entry = reader->getEntryFromTitle(*variantsItr);
+        entry = entry.getFinalEntry();
+        patternCorrespondingUrl = entry.getPath();
+        break;
+      } catch(kiwix::NoEntry& e) {
+        variantsItr++;
+      }
     }
 
     /* If article found then redirect directly to it */
     if (!patternCorrespondingUrl.empty()) {
       httpRedirection
-          = rootLocation + "/" + request_context->humanReadableBookId + "/" + patternCorrespondingUrl;
-      request_context->httpResponseCode = MHD_HTTP_FOUND;
+          = rootLocation + "/" + humanReadableBookId + "/" + patternCorrespondingUrl;
+      request->httpResponseCode = MHD_HTTP_FOUND;
       return build_response("", 0, httpRedirection, "", false, true);
     }
   }
 
   /* Make the search */
-  if (request_context->searcher != NULL) {
-    const char* start = MHD_lookup_connection_value(
-        request_context->connection, MHD_GET_ARGUMENT_KIND, "start");
-    const char* end = MHD_lookup_connection_value(
-        request_context->connection, MHD_GET_ARGUMENT_KIND, "end");
-    unsigned int startNumber = start != NULL ? atoi(start) : 0;
-    unsigned int endNumber = end != NULL ? atoi(end) : 25;
+  if (searcher != nullptr &&
+      (!patternString.empty() || has_geo_query)) {
+    auto start = 0;
+    try {
+      start = request->get_argument<unsigned int>("start");
+    } catch (const std::exception&) {}
+    auto end = 25;
+    try {
+      end = request->get_argument<unsigned int>("end");
+    } catch (const std::exception&) {}
 
     /* Get the results */
     pthread_mutex_lock(&searchLock);
     try {
-      request_context->searcher->search(patternString, startNumber, endNumber, isVerbose.load());
-      content = request_context->searcher->getHtml();
+      if (patternString.empty()) {
+        searcher->geo_search(latitude, longitude, distance,
+                             start, end, isVerbose.load());
+      } else {
+        searcher->search(patternString,
+                         start, end, isVerbose.load());
+      }
+      content = searcher->getHtml();
     } catch (const std::exception& e) {
       std::cerr << e.what() << std::endl;
     }
     pthread_mutex_unlock(&searchLock);
   } else {
     content = "<!DOCTYPE html>\n<html><head><meta content=\"text/html;charset=UTF-8\" http-equiv=\"content-type\" /><title>Fulltext search unavailable</title></head><body><h1>Not Found</h1><p>There is no article with the title <b>\"" + kiwix::encodeDiples(patternString) + "\"</b> and the fulltext search engine is not available for this content.</p></body></html>";
-    request_context->httpResponseCode = MHD_HTTP_NOT_FOUND;
+    request->httpResponseCode = MHD_HTTP_NOT_FOUND;
   }
 
   mimeType = "text/html; charset=utf-8";
 
-  introduceTaskbar(content, request_context->humanReadableBookId);
+  introduceTaskbar(content, humanReadableBookId);
 
-  bool deflated = request_context->acceptEncodingDeflate && compress_content(content, mimeType);
+  bool deflated = request->can_compress() && compress_content(content, mimeType);
   return build_response(content.data(),
                         content.size(),
                         httpRedirection,
                         mimeType,
                         deflated,
-                        true);
+                        cacheEnabled);
 }
 
-static struct MHD_Response* handle_random(RequestContext* request_context)
+static struct MHD_Response* handle_random(RequestContext* request)
 {
+  if (isVerbose.load()) {
+    printf("** running handle_random\n");
+  }
+
   std::string httpRedirection;
-  request_context->httpResponseCode = MHD_HTTP_FOUND;
-  if (request_context->reader != NULL) {
-    std::string randomUrl = request_context->reader->getRandomPageUrl();
+  request->httpResponseCode = MHD_HTTP_FOUND;
+  std::string humanReadableBookId;
+  try {
+    humanReadableBookId = request->get_argument("content");
+  } catch (const std::out_of_range&) {
+    return build_homepage(request);
+  }
+
+  auto reader = get_from_humanReadableBookId(humanReadableBookId).first;
+  if (reader == nullptr) {
+    return build_homepage(request);
+  }
+
+  try {
+    auto entry = reader->getRandomPage();
+    entry = entry.getFinalEntry();
     httpRedirection
-        = rootLocation + "/" + request_context->humanReadableBookId + "/" + kiwix::urlEncode(randomUrl);
+      = rootLocation + "/" + humanReadableBookId + "/" + kiwix::urlEncode(entry.getPath());
+    return build_response("", 0, httpRedirection, "", false, false);
+  } catch(kiwix::NoEntry& e) {
+    return build_404(request, humanReadableBookId);
   }
-  return build_response("", 0, httpRedirection, "", false, false);
 }
 
-static struct MHD_Response* handle_content(RequestContext* request_context)
+static struct MHD_Response* handle_catalog(RequestContext* request)
 {
+  if (isVerbose.load()) {
+    printf("** running handle_catalog");
+  }
+
+  std::string host;
+  std::string url;
+  try {
+    host = request->get_header("Host");
+    url  = request->get_url_part(1);
+  } catch (const std::out_of_range&) {
+    return build_404(request, "");
+  }
+
+  std::string content;
+  std::string mimeType;
+
+  if (url == "searchdescription.xml") {
+    content = catalogOpenSearchDescription;
+    mimeType = "application/opensearchdescription+xml";
+  } else {
+    zim::Uuid uuid;
+    kiwix::OPDSDumper opdsDumper;
+    opdsDumper.setRootLocation(rootLocation);
+    opdsDumper.setSearchDescriptionUrl("catalog/searchdescription.xml");
+    mimeType = "application/atom+xml;profile=opds-catalog;kind=acquisition; charset=utf-8";
+    kiwix::Library library_to_dump;
+    if (url == "root.xml") {
+      opdsDumper.setTitle("All zims");
+      uuid = zim::Uuid::generate(host);
+      library_to_dump = libraryManager.cloneLibrary();
+    } else if (url == "search") {
+      std::string query;
+      try {
+        query = request->get_argument("q");
+      } catch (const std::out_of_range&) {
+        return build_404(request, "");
+      }
+      opdsDumper.setTitle("Search result for " + query);
+      uuid = zim::Uuid::generate();
+      library_to_dump = libraryManager.filter(query);
+    } else {
+      return build_404(request, "");
+    }
+
+    {
+      std::stringstream ss;
+      ss << uuid;
+      opdsDumper.setId(ss.str());
+    }
+    opdsDumper.setLibrary(library_to_dump);
+    content = opdsDumper.dumpOPDSFeed();
+  }
+
+  bool deflated = request->can_compress() && compress_content(content, mimeType);
+  return build_response(
+      content.data(), content.size(), "", mimeType, deflated, false);
+}
+
+static struct MHD_Response* handle_content(RequestContext* request)
+{
+  if (isVerbose.load()) {
+    printf("** running handle_content\n");
+  }
+
   std::string baseUrl;
   std::string content;
   std::string mimeType;
 
-  bool found = false;
-  zim::Article article;
+  kiwix::Entry entry;
+
+  std::string humanReadableBookId;
   try {
-    found = request_context->reader->getArticleObjectByDecodedUrl(request_context->urlStr, article);
-
-    if (found) {
-      /* If redirect */
-      unsigned int loopCounter = 0;
-      while (article.isRedirect() && loopCounter++ < 42) {
-        article = article.getRedirectArticle();
-      }
-
-      /* To many loop */
-      if (loopCounter == 42)
-        found = false;
-    }
-  } catch (const std::exception& e) {
-    std::cerr << e.what() << std::endl;
-    found = false;
+    humanReadableBookId = request->get_url_part(0);
+  } catch (const std::out_of_range& e) {
+    return build_homepage(request);
   }
 
-  if (!found) {
+  auto reader = get_from_humanReadableBookId(humanReadableBookId).first;
+  if (reader == nullptr) {
+    return build_homepage(request);
+  }
+
+  auto urlStr = request->get_url().substr(humanReadableBookId.size()+1);
+
+  try {
+    entry = reader->getEntryFromPath(urlStr);
+    entry = entry.getFinalEntry();
+  } catch(kiwix::NoEntry& e) {
     if (isVerbose.load())
-      printf("Failed to find %s\n", request_context->urlStr.c_str());
+      printf("Failed to find %s\n", urlStr.c_str());
 
-    return build_404(request_context);
+    return build_404(request, humanReadableBookId);
   }
 
   try {
-    mimeType = article.getMimeType();
+    mimeType = entry.getMimetype();
   } catch (exception& e) {
     mimeType = "application/octet-stream";
   }
 
   if (isVerbose.load()) {
-    printf("Found %s\n", request_context->urlStr.c_str());
+    printf("Found %s\n", urlStr.c_str());
     printf("mimeType: %s\n", mimeType.c_str());
   }
 
   if (mimeType.find("text/") != string::npos
       || mimeType.find("application/javascript") != string::npos
       || mimeType.find("application/json") != string::npos) {
-    zim::Blob raw_content = article.getData();
+    zim::Blob raw_content = entry.getBlob();
     content = string(raw_content.data(), raw_content.size());
 
     /* Special rewrite URL in case of ZIM file use intern *asbolute* url like
      * /A/Kiwix */
     if (mimeType.find("text/html") != string::npos) {
-      baseUrl = "/" + std::string(1, article.getNamespace()) + "/"
-                + article.getUrl();
+      baseUrl = "/" + entry.getPath();
       pthread_mutex_lock(&regexLock);
       content = replaceRegex(content,
-                             "$1$2" + rootLocation + "/" + request_context->humanReadableBookId + "/$3/",
+                             "$1$2" + rootLocation + "/" + humanReadableBookId + "/$3/",
                              "(href|src)(=[\"|\']{0,1})/([A-Z|\\-])/");
       content = replaceRegex(content,
-                             "$1$2" + rootLocation + "/" + request_context->humanReadableBookId + "/$3/",
+                             "$1$2" + rootLocation + "/" + humanReadableBookId + "/$3/",
                              "(@import[ ]+)([\"|\']{0,1})/([A-Z|\\-])/");
       content = replaceRegex(
           content,
-          "<head><base href=\"" + rootLocation + "/" + request_context->humanReadableBookId + baseUrl + "\" />",
+          "<head><base href=\"" + rootLocation + "/" + humanReadableBookId + baseUrl + "\" />",
           "<head>");
       pthread_mutex_unlock(&regexLock);
-      introduceTaskbar(content, request_context->humanReadableBookId);
+      introduceTaskbar(content, humanReadableBookId);
     } else if (mimeType.find("text/css") != string::npos) {
       pthread_mutex_lock(&regexLock);
       content = replaceRegex(content,
-                             "$1$2" + rootLocation + "/" + request_context->humanReadableBookId + "/$3/",
+                             "$1$2" + rootLocation + "/" + humanReadableBookId + "/$3/",
                              "(url|URL)(\\([\"|\']{0,1})/([A-Z|\\-])/");
       pthread_mutex_unlock(&regexLock);
     }
 
     bool deflated
-        = request_context->acceptEncodingDeflate && compress_content(content, mimeType);
+        = request->can_compress() && compress_content(content, mimeType);
     return build_response(
         content.data(), content.size(), "", mimeType, deflated, true);
   } else {
     int range_len;
-    if (request_context->range_end == -1) {
-       range_len = article.getArticleSize() - request_context->range_start;
+    if (request->get_range().second == -1) {
+       range_len = entry.getSize() - request->get_range().first;
     } else {
-       range_len = request_context->range_end - request_context->range_start;
+       range_len = request->get_range().second - request->get_range().first;
     }
-    return build_callback_response_from_article(
-      article,
-      request_context->range_start,
+    return build_callback_response_from_entry(
+      entry,
+      request->get_range().first,
       range_len,
       mimeType);
   }
 }
 
-static struct MHD_Response* handle_default(RequestContext* request_context)
-{
-  std::string content = welcomeHTML;
-
-  std::string mimeType = "text/html; charset=utf-8";
-
-  bool deflated = request_context->acceptEncodingDeflate && compress_content(content, mimeType);
-  return build_response(
-      content.data(), content.size(), "", mimeType, deflated, false);
-}
-
-int print_out_key (void *cls, enum MHD_ValueKind kind,
+int print_key_value (void *cls, enum MHD_ValueKind kind,
                    const char *key, const char *value)
 {
-  printf ("%s: %s\n", key, value);
+  printf (" - %s: '%s'\n", key, value);
   return MHD_YES;
 }
 
@@ -645,135 +820,56 @@ static int accessHandlerCallback(void* cls,
                                  size_t* upload_data_size,
                                  void** ptr)
 {
-  /* Unexpected method */
-  if (0 != strcmp(method, "GET") && 0 != strcmp(method, "POST"))
-    return MHD_NO;
-
-  if (isVerbose.load()) {
+  if (isVerbose.load() ) {
+    printf("======================\n");
     printf("Requesting : \n");
-    printf("u : %s\n", url);
-    printf("m : %s\n", method);
-    printf("v : %s\n", version);
-    MHD_get_connection_values (connection, MHD_HEADER_KIND, &print_out_key, NULL);
+    printf("full_url  : %s\n", url);
   }
+  RequestContext request(connection, rootLocation, url, method, version);
 
-  /* Check if the response can be compressed */
-  const char* acceptEncodingHeaderValue = MHD_lookup_connection_value(
-      connection, MHD_HEADER_KIND, MHD_HTTP_HEADER_ACCEPT_ENCODING);
-  const bool acceptEncodingDeflate
-      = acceptEncodingHeaderValue
-        && string(acceptEncodingHeaderValue).find("deflate") != string::npos;
-
-  /* Check if range is requested. */
-  const char* acceptRangeHeaderValue = MHD_lookup_connection_value(
-      connection, MHD_HEADER_KIND, MHD_HTTP_HEADER_RANGE);
-  bool acceptRange = false;
-  int range_start = 0;
-  int range_end = -1;
-  if (acceptRangeHeaderValue != NULL) {
-    auto range = std::string(acceptRangeHeaderValue);
-    if (range.substr(0, 6) == "bytes=")
-    {
-      range = range.substr(6);
-      std::istringstream iss(range);
-      iss >> range_start;
-      range = range.substr(iss.tellg());
-      if (range[0] == '-') {
-        range = range.substr(1);
-        if (! range.empty()) {
-          std::istringstream iss(range);
-          iss >> range_end;
-          range = range.substr(iss.tellg());
-        }
-        if (range.empty()) {
-          // Nothing left to read. We are OK.
-          acceptRange = true;
-        }
-      }
-    }
+  if (isVerbose.load() ) {
+    request.print_debug_info();
+  }
+  /* Unexpected method */
+  if (request.get_method() != RequestMethod::GET && request.get_method() != RequestMethod::POST) {
+    printf("Reject request because of unhandled request method.\n");
+    printf("----------------------\n");
+    return MHD_NO;
   }
 
   /* Prepare the variables */
   struct MHD_Response* response;
-  int httpResponseCode = acceptRange ? MHD_HTTP_PARTIAL_CONTENT : MHD_HTTP_OK;
-  std::string urlStr = string(url);
+  request.httpResponseCode = request.has_range() ? MHD_HTTP_PARTIAL_CONTENT : MHD_HTTP_OK;
 
-  /* Get searcher and reader */
-  std::string humanReadableBookId = "";
-
-  if (!rootLocation.empty() && urlStr.substr(0, rootLocation.size() + 1) != rootLocation + "/"){
-        humanReadableBookId = "";
-  }
-
-  else if (!(urlStr.size() > rootLocation.size() + 5 && urlStr.substr(rootLocation.size() , 6) == "/skin/")) {
-    if ((urlStr == rootLocation + "/" + "search") || (urlStr == rootLocation + "/" + "suggest")
-        || (urlStr == rootLocation + "/" + "random")) {
-      const char* tmpGetValue = MHD_lookup_connection_value(
-          connection, MHD_GET_ARGUMENT_KIND, "content");
-      humanReadableBookId = (tmpGetValue != NULL ? string(tmpGetValue) : "");
+  if (! request.is_valid_url()) {
+    response = build_homepage(&request);
+  } else {
+    if (startswith(request.get_url(), "/skin/")) {
+      response = handle_skin(&request);
+    } else if (startswith(request.get_url(), "/catalog")) {
+      response = handle_catalog(&request);
+    } else if (request.get_url() == "/meta") {
+      response = handle_meta(&request);
+    } else if (request.get_url() == "/search") {
+      response = handle_search(&request);
+    } else if (request.get_url() == "/suggest") {
+      response = handle_suggest(&request);
+    } else if (request.get_url() == "/random") {
+      response = handle_random(&request);
     } else {
-      humanReadableBookId = urlStr.substr(rootLocation.size() + 1,
-                                          urlStr.find("/", rootLocation.size() + 1) != string::npos
-                                              ? urlStr.find("/", rootLocation.size() + 1) -  (rootLocation.size() + 1)
-                                              : urlStr.size() - (rootLocation.size() + 2));
-      if (!humanReadableBookId.empty()) {
-        urlStr = urlStr.substr(urlStr.find("/", rootLocation.size() + 1) != string::npos
-                                   ? urlStr.find("/", rootLocation.size() + 1)
-                                   : humanReadableBookId.size());
-      }
+      response = handle_content(&request);
     }
   }
 
-  kiwix::Searcher* searcher
-      = searchers.find(humanReadableBookId) != searchers.end()
-            ? searchers.find(humanReadableBookId)->second
-            : globalSearcher;
-  kiwix::Reader* reader = readers.find(humanReadableBookId) != readers.end()
-                              ? readers.find(humanReadableBookId)->second
-                              : NULL;
-  if (reader == NULL) {
-    humanReadableBookId = "";
-  }
-
-  RequestContext request_context(connection, httpResponseCode,
-                                 reader, searcher,
-                                 urlStr, humanReadableBookId,
-                                 acceptEncodingDeflate,
-                                 acceptRange, range_start, range_end);
-
-
-  /* Get suggestions */
-  if ((urlStr == (rootLocation + "/" + "suggest")) && reader != NULL) {
-    response = handle_suggest(&request_context);
-  }
-
-  /* Get static skin stuff */
-  else if (urlStr.size() > rootLocation.size() + 5 && urlStr.substr(rootLocation.size() , 6) == "/skin/") {
-    response = handle_skin(&request_context);
-  }
-
-  /* Display the search restults */
-  else if (urlStr == (rootLocation + "/" + "search")) {
-    response = handle_search(&request_context);
-  }
-
-  /* Display a random article */
-  else if (urlStr == (rootLocation + "/" + "random")) {
-    response = handle_random(&request_context);
-  }
-
-  /* Display the content of a ZIM content (article, image, ...) */
-  else if (reader != NULL) {
-    response = handle_content(&request_context);
-  }
-
-  /* Display the global Welcome page */
-  else {
-    response = handle_default(&request_context);
-  }
-
   /* Queue the response */
-  int ret = MHD_queue_response(connection, request_context.httpResponseCode, response);
+  if (isVerbose.load()) {
+    printf("Response :\n");
+    printf("httpResponseCode : %d\n", request.httpResponseCode);
+    printf("headers :\n");
+    MHD_get_response_headers(response, print_key_value, nullptr);
+    printf("----------------------\n");
+  }
+  int ret = MHD_queue_response(connection, request.httpResponseCode, response);
   MHD_destroy_response(response);
 
   return ret;
@@ -788,12 +884,11 @@ int main(int argc, char** argv)
   string rootPath;
   string interface;
   int serverPort = 80;
-  int daemonFlag = false;
+  int daemonFlag [[gnu::unused]] = false;
   int libraryFlag = false;
   string PPIDString;
   unsigned int PPID = 0;
   unsigned int nb_threads = std::thread::hardware_concurrency();
-  kiwix::Manager libraryManager;
 
   static struct option long_options[]
       = {{"daemon", no_argument, 0, 'd'},
@@ -912,9 +1007,7 @@ int main(int argc, char** argv)
         try {
           string libraryPath
               = isRelativePath(*itr)
-                    ? computeAbsolutePath(removeLastPathElement(
-                                              getExecutablePath(), true, false),
-                                          *itr)
+                    ? computeAbsolutePath(getCurrentDirectory(), *itr)
                     : *itr;
           retVal = libraryManager.readFile(libraryPath, true);
         } catch (...) {
@@ -977,7 +1070,7 @@ int main(int argc, char** argv)
         readers[humanReadableId] = reader;
 
         if ( reader->hasFulltextIndex()) {
-          kiwix::Searcher* searcher = new kiwix::Searcher();
+          kiwix::Searcher* searcher = new kiwix::Searcher(humanReadableId);
           searcher->setProtocolPrefix(rootLocation + "/");
           searcher->setSearchProtocolPrefix(rootLocation + "/" + "search?");
           searcher->add_reader(reader, humanReadableId);
@@ -1035,6 +1128,11 @@ int main(int argc, char** argv)
 
   introduceTaskbar(welcomeHTML, "");
 
+  /* Compute the OpenSearch description */
+  catalogOpenSearchDescription = RESOURCE::opensearchdescription_xml;
+  catalogOpenSearchDescription = replaceRegex(catalogOpenSearchDescription, rootLocation, "__ROOT_LOCATION__");
+
+
 #ifndef _WIN32
   /* Fork if necessary */
   if (daemonFlag) {
@@ -1154,7 +1252,7 @@ int main(int argc, char** argv)
   }
 
   if (daemon == NULL) {
-    cerr << "Unable to instanciate the HTTP daemon. The port " << serverPort
+    cerr << "Unable to instantiate the HTTP daemon. The port " << serverPort
          << " is maybe already occupied or need more permissions to be open. "
             "Please try as root or with a port number higher or equal to 1024."
          << endl;
